/*
 * 
 * main.c (Mainloop of lpt_driver)
 * 
 * An Linux LPT user-space driver for hacking with external LCD displays
 * and DS1620 Temp Sensors.
 * 
 * 1 Feb 98 to 10 July 98     Version 0.30
 * ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** 
 * Copyright (c) 1998 Carl D. Walker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** *****
 *
 * See the file 'lpt_driver_doc.html' for information on how to connect the
 * lpt port to the display and the temp sensor.
 * 
 * NOTE: Run as root. This is needed for the ioperm() calls to work
 *       (start it up from rc.local, if you wish...)
 */
/* From version 0.29 to version 0.30, fixed a bug in the display driver
 * code that could cause a corrupt display, thanks to Tony Griggs
 * for this one.
 */

#define _POSIX_SOURCE 1

#include <linux/unistd.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/lp.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <string.h>
#include <syslog.h>
#include <stdlib.h>

#define BUSY_READ_MASK 0x80
#define DUMMY_OP_PORT 0x80
#define LP1_BASE 0x278
#define LP0_BASE 0x378
#define PORT_BASE LP0_BASE
#define DALLAS_READ_TEMP 0xaa
#define DEGREES_F 0
#define DEGREES_C 1
#define TEMP_NOT_VALID -99

#include "lpt_driver.h"

#include "misc.c"
#include "io.c"
#include "fifo.c"
#include "dallas.c"

/*
 *  ***** Globals ***** 
 */

/* become a daemon if set (using 0 here is useful for debug)*/
unsigned char daemon = 1;   
time_t clock_time;
char scratch_string[81] = "";
char time_message[81] = "";
char *pointer;
char *time_message_pointer;
char *ctime_pointer;

/* String space for display */
char string_data[81] = " ";
char clock_string[81] = " ";
char uptime_string[81] = " ";
char textload_string[81] = " ";
char graphload_string[81] = " ";
char min_string[81] = " ";
char max_string[81] = " ";
char pid_string[81] = " ";

int my_pid;
int sysinfo_stat;
int temp_time_counter;
float version;
float uptime_hours;
float uptime_days;
float load1;
float load5;
float load15;
int procs;
int i;	
char offset_string[] = "-offset\0";
/* our home grown degree symbol */
char deg_str[] = "\006\0";		 
/* our arrowhead symbol */
char arrow_str[] = "\007\0";	
/* Default display mode */
int display_mode;			 

/*
 * ***** The main loop of the example driver ***** 
 */
int main(int argc, char *argv[]) {
	
/*
 *  ***** Command line interface ***** 
 */
	disp_size = e_24x2;                    
	/* Set default display size (see header file!) */
	display_mode = CLOCK_PROCS;         
	/* Set default display mode */
	dal_dat.offset = 0;			 
	/* temp offset 0 unless overridden */
	dal_dat.temp_display_type = DEGREES_F; 
	/* default temp display mode */
	fifo.init_ok = 0;			 
	/* FIFO not init'ed yet */
	disp_cp.auto_fill = 1;		 
	/* fill displayed lines to the end */
	/* with spaces */
	disp_cp.fill_data = 0x20;		 
	/* set the rev level */
	version = 0.30;			 
	temp_time_counter = 1;
	
	/* decipher the command line switches, */
	/* with some very basic error checking */
	if (argc > 1) {
	  for (i=1; i<argc; i++) { 
	    /* get temp sensor offset in deg_c */
	    if ((strcmp((char*)argv[i], (char*)&offset_string)) == 0) {
	      if (i == (argc - 1)) {	
		printf("Command-line Error: -offset switch must have an argument.\n");
		exit(-1); 
		/* bail-out if the command line is broken */
	      }
	      dal_dat.offset = atof((char*)argv[(i+1)]);
	      printf("Offset enabled, %3.2f Degrees C\n",
		     dal_dat.offset);		
	    }
	  }
	}
	
	time_message_pointer = (char *)(&time_message); 
	/* init all things that need initing */
	dal_dat.protocol = DALLAS_READ_TEMP;	
	/* set the value for the read_temp protocol */
    
	if (daemon) {
	  my_pid =  daemon_init(version); 	
	  /* break away and go into daemon mode */
	}
	
	/* Make a string for display with the current PID in it */
	sprintf(&pid_string,"PID: %5d Ver. %1.2f",my_pid,version);
	
	/* Handle shutdown and other signals cleanly */
	signal(SIGINT, clean_up);
	signal(SIGTERM, clean_up);
	signal(SIGHUP, alt_temp);             
	/* Change temp display mode */
	signal(SIGUSR1, alt_mode);  
	/* Change the display mode */
	signal(SIGUSR2, log_temp);  
	/* Log the temp to the syslog in whatever mode (C|F) */
	/* we are displaying at the time */
	
	syslog_open();         		
	/* open the syslog */
	sprintf(syslog_string_ptr,
		"LCD display driver daemon V%1.2f activated.\n",
		version);
	syslog_send(syslog_string_ptr);	
	disp_init(port_base_addr,port_len);
	/* allocate the needed IO ports */
	
	time(&clock_time); 
	/* get the time */
	/* convert time to ascii message */
	ctime_pointer = (char *)ctime(&clock_time);
	/* copy the string to a place we can work on it */
	strcpy(time_message_pointer,ctime_pointer);
	
	dallas_init();
	/* probe/setup dallas device */  
	init_fifo();	
	/* check the plumbing */
	
	sleep(1);
	pointer = (char *)(&string_data);	
	/* display a logon message */
	
	switch (disp_size) {
	case e_20x2:
	  sprintf(pointer,
		  "lpt_driver V%1.2f GPL",
		  version); 
	  /* 1st logon message */
	  display_address(0x00);
	  display_string(pointer);
	  line_2();
	  sprintf(pointer,
		  " ICMP Networks 1998 ");
	  display_string(pointer);
	  sleep(3);
	  break;
	  
	case e_24x2:
	  sprintf(pointer,
		  "  lpt_driver V%1.2f GPL",
		  version); 
	  /* 1st logon message */
	  display_address(0x00);
	  display_string(pointer);
	  line_2();
	  sprintf(pointer,
		  "   ICMP Networks 1998 ");
	  display_string(pointer);
	  sleep(3);
	  break;
	} /* end switch (disp_size)... */
	
	/* The real MainLoop */	
	
	while (1) {
	  /* read the temp from the dallas device and see if it is there */
	  if (dal_dat.sensor_ok) {
	    temp_time_counter--;
	    if (temp_time_counter <= 0) {
	      get_temp(DEGREES_F);		
	      /* get the temp in deg_f */
	      temp_time_counter = 60;	
	      /* temp delay time in seconds */	
	    }
	  }
	  /* get the system information */
	  sysinfo_stat = sysinfo(sysin_ptr);
	  if (sysinfo_stat == -1) {		
	    /* some error on sysinfo call */
	    exit(-1);
	  }
	  
	  uptime_hours = (((sysin.uptime) / 60.0) / 60.0);
	  uptime_days = (uptime_hours / 24.0); 
	  /* calculate system up-time */
	  procs = sysin.procs;
	  /* number of system procs */
	  load1  = sysin.loads[0] / 64000.0;
	  /* mangle the system load data into */
	  /* a more printable form */
	  load5  = sysin.loads[1] / 64000.0;	 
	  load15 = sysin.loads[2] / 64000.0;
	  
	  switch (disp_size) {      
	  case e_20x2:
	    if (dal_dat.minus) {    
	      /* Shorten bar graph length to 8 chars */
	      /* if the temp is negitive */
	      calc_bar_g_display(load1,8);    
	      /* This fix bumps the rev from v22 to v23 */
	      /* 12 April 98 CW */
	    }
	    else {
	      calc_bar_g_display(load1,9);	
	      /* build a bar-graph */		
	    }
	    break;
	    
	  case e_24x2:
	    if (dal_dat.minus) {
	      calc_bar_g_display(load1,12);
	      /* a 2 line x 24 char display..... */
	    }
	    else {
	      calc_bar_g_display(load1,13);
	    }	  
	    break;
	    
	  }/* end switch (disp_size)... */
	  
	  display_address(0x00);		
	  /* point at the upper left of the display */
	  pointer = (char *)(&string_data);	
	  check_and_display_fifo();
	  /* display fifo data if there is */
	  if (fifo.update > 1) {	
	    /* see if a display has timed-out */
	    /* fifo.update > zero_based_message_display_delay_in_seconds */	
	    /* build the display strings for all the display modes */
	    if (disp_size != e_20x2) {	
	      /* display line one for 24x2 display */
	      if (uptime_hours < 1000) {
		sprintf(&uptime_string,"Up: %4.2f Hrs Proc:%3d",
			uptime_hours,
			procs);
	      }
	      else {
		sprintf(&uptime_string,"Up:%4.2f Hrs Proc:%3d",
			uptime_hours,
			procs);
	      }
	    }
	    
	    else {				
	      /* must be a 20x2 display */
	      if (uptime_hours < 1000) {
		sprintf(&uptime_string,
			"Up %3.2fhr Proc%3d",
			uptime_hours,
			procs);
	      }
	      else {
		sprintf(&uptime_string,
			"Up%3.2fhr Proc%3d",
			uptime_hours,
			procs);
	      }
	    }
	    
	    if ((dal_dat.sensor_ok) & 
		(dal_dat.temp_display_type == DEGREES_F)) {
	      sprintf(&graphload_string,"%3.1f%sF -%s%s",\
		      dal_dat.degrees_f,
		      deg_str,
		      arrow_str,
		      bar_g.display_string);
	      sprintf(&textload_string,
		      "%3.1fF Load 1:%1.2f 5:%1.2f",
		      dal_dat.degrees_f,
		      load1,
		      load5);
	      sprintf(&min_string,
		      "%3.1f%sF Minimum: %3.1f%sF",
		      dal_dat.degrees_f,
		      deg_str,
		      dal_dat.min_f,
		      deg_str);
	      sprintf(&max_string,
		      "%3.1f%sF Maximum: %3.1f%sF",
		      dal_dat.degrees_f,
		      deg_str,
		      dal_dat.max_f,
		      deg_str);
	    }
	    
	    else if ((dal_dat.sensor_ok) & 
		     (dal_dat.temp_display_type == DEGREES_C)) {
	      sprintf(&graphload_string,
		      "%3.1f%sC -%s%s",
		      dal_dat.degrees_c,
		      deg_str,
		      arrow_str,
		      bar_g.display_string);
	      sprintf(&textload_string,
		      "%3.1fC Load 1:%1.2f 5:%1.2f",
		      dal_dat.degrees_c,
		      load1,
		      load5);
	      sprintf(&min_string,
		      "%3.1f%sC Minimum: %3.1f%sC",
		      dal_dat.degrees_c,
		      deg_str,
		      dal_dat.min_c,
		      deg_str);
	      sprintf(&max_string,
		      "%3.1f%sC Maximum: %3.1f%sC",
		      dal_dat.degrees_c,
		      deg_str,
		      dal_dat.max_c,
		      deg_str);
	    }
	    
	    else {
	      sprintf(&graphload_string,
		      "NO TEMP  %s",
		      bar_g.display_string);
	      sprintf(&textload_string,
		      "Load 1:%2.2f 5:%2.2f 15:%2.2f",
		      load1,
		      load5,
		      load15);
	    }
	    
	    /* set cursor to the first line */
	    display_address(0x00);
	    
	    /* select the strings to display based on the */
	    /* current display mode */
	    /* from the display selection state machine */
	    switch (display_mode) {
	    case PROCS:
	      pointer = (unsigned char*)&uptime_string;
	      display_string(pointer);
	      line_2();
	      pointer = (unsigned char*)&graphload_string;
	      display_string(pointer);
	      break;
	      
	    case CLOCK:
	      display_time();
	      line_2();
	      pointer = (unsigned char*)&graphload_string;
	      display_string(pointer);
	      break;
	      
	    case PROCS_TEXTLOAD:
	      pointer = (unsigned char*)&uptime_string;
	      display_string(pointer);
	      line_2();
	      pointer = (unsigned char*)&textload_string;
	      display_string(pointer);
	      break;
	      
	    case CLOCK_TEXTLOAD:
	      display_time();
	      line_2();
	      pointer = (unsigned char*)&textload_string;
	      display_string(pointer);
	      break;
	      
	    case MIN_TEMP:
	      display_time();
	      line_2();
	      pointer = (unsigned char*)&min_string;
	      display_string(pointer);
	      break;
	      
	    case MAX_TEMP:
	      display_time();
	      line_2();
	      pointer = (unsigned char*)&max_string;
	      display_string(pointer);
	      break;
	      
	    case MY_PID:
	      display_time();
	      line_2();
	      pointer = (unsigned char*)&pid_string;
	      display_string(pointer);
	      break;
	      
	    case CLOCK_PROCS:
	      display_time();
	      line_2();
	      pointer = (unsigned char*)&uptime_string;
	      display_string(pointer);
	      break;
	    } /* end switch(display_mode)... */
	    
	    pointer = (char *)(&string_data); 
	    /* set the string pointer back to the default scratch string */
	    
	  } /* END if (fifo.update..... */
	  
	wait_point:
	  sleep (1);
	}	
}

void display_time(void) {
  time(&clock_time); 
  /* get the time */
  /* convert time to ascii message */
  ctime_pointer = (char *)ctime(&clock_time);
  /* copy the string to a place we can work on it */
  strcpy(time_message_pointer,ctime_pointer);
  
  if (disp_size == e_20x2)
    (time_message_pointer[19]) = 0x00; 
  /* strip off the year and everything */
  /* after it by inserting a string terminator */
  else
    (time_message_pointer[24]) = 0x00;  
  
  display_string((char *)time_message_pointer); 
  /* send it out */
}

void alt_mode(void) {
  /* state machine to select display operating mode */
  switch (display_mode) {
  case PROCS: 
    display_mode = CLOCK;
    break;
  case CLOCK:
    display_mode = PROCS_TEXTLOAD;
    break;
  case PROCS_TEXTLOAD:
    display_mode = CLOCK_TEXTLOAD;
    break;
  case CLOCK_TEXTLOAD:
    if (!dal_dat.sensor_ok)
      /* Don't display temp. information if the sensor is no good */
      display_mode = MY_PID;
    else
      display_mode = MIN_TEMP;
    break;
  case MIN_TEMP:
    display_mode = MAX_TEMP;
    break;
  case MAX_TEMP:
    display_mode = MY_PID;
    break;
  case MY_PID:
    display_mode = CLOCK_PROCS;
    break;
  case CLOCK_PROCS:
    display_mode = PROCS;
    break;
  }
}
